// Class: TraversalGUI
//
// Author: Alyce Brady
//
// License Information:
//   This class is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation.
//
//   This class is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 *  Environment-Based Applications:<br>
 *
 *    The TraversalGUI class provides a graphical user interface to
 *    a program that creates an animation of 2-D structure traversals.
 *
 *  @author Alyce Brady
 *  @version 1 September 2002
 **/

public class TraversalGUI extends JFrame
{

    ///////////////////////////////////////////////////////////////////////
    // Static Data: Variables and Constants not tied to any one instance //
    ///////////////////////////////////////////////////////////////////////

    // variables to number windows
    private static int windowCount = 0;      // how many have been created?
    private static int openWindowCount = 0;  // how many are now open?

    // list of traversal iterators; algNameList is a list of iterator
    // names in the order in which they are registered, while iteratorMap
    // maps those names to EnvIterator or one of its subclasses
    private static ArrayList algNameList = new ArrayList();
    private static Map iteratorMap = new HashMap();

    // constants used to initialize slider
    private static final int MIN_DELAY_MSECS = 10;
    private static final int MAX_DELAY_MSECS = 1000;
    private static final int INITIAL_DELAY = MIN_DELAY_MSECS +
                                    (MAX_DELAY_MSECS-MIN_DELAY_MSECS)/2;

    // iterator constructor parameter types (EnvIterator and subclasses)
    private static final Class[] IT_CONST_PARAM_TYPES = {BoundedEnv.class};

    /////////////////////////////////////////////////////////
    // Instance Variables for a TraversalGUI window object //
    /////////////////////////////////////////////////////////

    private BoundedEnv theEnv;          // environment to traverse
    private int numLocationsInEnv;      // number of locations in theEnv

    private ScrollableEnvDisplay display;  // object to display the environment
    private int displayWidth;           // width of display area in pixels
    private int displayHeight;          // height of display area in pixels

    private EnvIterator it;             // iterator for traversing theEnv
    private int numLocationsVisited;    // number of locs visited in traversal
    private Color bgColor;              // background color for environment
    private Color highlightColor;       // to highlight cells while traversing
    private Timer timer;                // timer controls animation speed

    private JPanel algButtonList;       // tool bar holding algorithm buttons
    private JButton stopButton, resetButton, newWindowButton; // control buttons


    //////////////////////////////////////////////////
    // Class Methods to Register and Find Iterators //
    //////////////////////////////////////////////////

    public static void register(String name, Class iteratorClass)
    {
        if ( ! EnvIterator.class.isAssignableFrom(iteratorClass) )
            throw new IllegalArgumentException("iterator " + 
              iteratorClass.getName() + " must be a subclass of EnvIterator");
        algNameList.add(name);
        iteratorMap.put(name, iteratorClass);
    }

    /** Construct the iterator with the specified name.
     *      @param  name name (label) of a traversal algorithm
     *      @return      an iterator that traverses in the specified order
     **/
    public static EnvIterator getIterator(String name, BoundedEnv env)
    {
        Class itClass = (Class) iteratorMap.get(name);
        try
        {
            Class[] pTypes = new Class[] {BoundedEnv.class};
            Constructor itCons = itClass.getConstructor(pTypes);
            Object[] params = new Object[] {env};
            return (EnvIterator) itCons.newInstance(params);
        }
        catch (Exception e)
        {  throw new MBSException("Error constructing " + itClass + 
            " iterator: " + e);
        }
    }


    /////////////////
    // Constructor //
    /////////////////

    /** Construct the graphical user interface.
     *  (Precondition: env and highlightColor are not null.)
     *  @param env     bounded environment in which to animate traversals
     *  @param backgroundColor background color for environment
     *  @param highlightColor  highlight color to illustrate traversal
     *  @param displayWidth    width of visible display area in pixels
     *  @param displayHeight   height of visible display area in pixels
     **/
    public TraversalGUI(BoundedEnv env, Color backgroundColor,
                        Color highlightColor,
                        int displayWidth, int displayHeight)
    {
        this.theEnv = env;
        this.numLocationsInEnv = theEnv.numRows() * theEnv.numCols();
        this.displayWidth = displayWidth;
        this.displayHeight = displayHeight;
        this.it = null;
        this.numLocationsVisited = 0;
        this.bgColor = backgroundColor;
        this.highlightColor = highlightColor;
        this.timer = new Timer(INITIAL_DELAY, new ActionListener() {
            public void actionPerformed(ActionEvent evt) { step(); }});

        // Reset exception handler to notify user of exception and
        // generate trace information.
        System.setProperty("sun.awt.exception.handler",
                           GUIExceptionHandler.class.getName());
 
        // Set characteristics for the main frame.
        windowCount++;
        openWindowCount++;
        String title = "Animating 2-D Traversals";
        if ( windowCount > 1 )
            title += " " + windowCount;
        setTitle(title);
        int offset = 50 * (windowCount - 1);
        setLocation(25 + offset , 15 + offset);
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) { closingWindow(); }});

        // Create a panel that will contain everything in the window.
        JPanel content = createControlPanel(); //new JPanel();
        setContentPane(content);
        content.setBorder(BorderFactory.createEmptyBorder(15, 15, 15, 15));
        content.setLayout(new BorderLayout());

        // Create a control panel on the left, a panel with the
        // environment display and a speed slider on the right,
        // and add them to the main panel.
        content.add(createControlPanel(), BorderLayout.CENTER);
        content.add(createGraphicsPanel(), BorderLayout.EAST);

        // Controls are enabled except while an algorithm is running.
        enableControls(true);

        pack();

        // Set the environment for the display.  Doing this AFTER packing
        // the contents into the window means that the display can be
        // smarter about determining the size of the environment cells.
        display.setEnv(theEnv);
    }

    /////////////////////////////////////////
    // Helper Methods for Constructing GUI //
    /////////////////////////////////////////

    /** Create the panel for displaying the control buttons.
     */
    private JPanel createControlPanel()
    {
        // Construct the panel in which to put all control buttons.
        // Algorithm buttons go on top, non-algorithm buttons on
        // bottom, and a blank panel to separate them goes in between.
        JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout());
        panel.add(new JPanel(), BorderLayout.CENTER);

        // Create buttons for the algorithms in the algorithm list.
        panel.add(createTraversalAlgButtons(), BorderLayout.NORTH);

        // Create buttons not tied to specific algorithms.
        panel.add(createAlgIndependentButtons(), BorderLayout.SOUTH);

        return panel;
    }

    /** Create a panel containing buttons that run traversal algorithms.
     */
    private JPanel createTraversalAlgButtons()
    {
        // Construct a labeled panel for the traversal algorithm buttons.
        // The panel is an instance variable so enableControls can access it.
        algButtonList = new JPanel();
        algButtonList.setLayout(new GridLayout(0, 1));
        algButtonList.setBorder(BorderFactory.createTitledBorder("Traversal Algorithms"));

        // For each algorithm in algNameList, make an action button.
        for ( int i = 0; i < algNameList.size(); i++ )
        {
            // Make a button that runs this traversal algorithm.
            String name = (String) algNameList.get(i);
            JButton algButton;
            algButtonList.add(algButton = new JButton(name));
            algButton.addActionListener(new AbstractAction(name) {
                public void actionPerformed(ActionEvent e)
                    { runTraversal((String) getValue(NAME)); } });
        }

        return algButtonList;
    }

    /** Create a panel containing buttons not tied to any one algorithm.
     */
    private JPanel createAlgIndependentButtons()
    {
        // Construct the panel in which to put other control buttons.
        // The buttons are instance variables so enableControls can access it.
        JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout());

        // Create stop and reset buttons and group them.
        JPanel subPanel = new JPanel();
        subPanel.add(stopButton = new JButton("Stop"));
        subPanel.add(resetButton = new JButton("Reset"));
        panel.add(subPanel, BorderLayout.CENTER);

        // Create button that creates a new window.
        panel.add(newWindowButton = new JButton("Open a New Traversal Window"),
                  BorderLayout.SOUTH);

        // Define the actions to associate with the buttons.
        stopButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) { stopAlg(); }});
        resetButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) { resetEnv(); }});
        newWindowButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) { newWindow(); }});

        return panel;
    }

    /** Create the panel for displaying the algorithm animation.
     */
    private JPanel createGraphicsPanel()
    {
        // Construct the panel in which to put display and slider.
        JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout());

        // Construct the graphics environment display, specifying
        // the dimensions of its visible area.  Create a scrollable
        // pane with a viewport that views the environment display.
        // Add it to the overall display panel.
        display = new ScrollableEnvDisplay(displayWidth, displayHeight, bgColor);
        JScrollPane scrollPane = new JScrollPane();
        PseudoInfiniteViewport vp = new PseudoInfiniteViewport(scrollPane);
        scrollPane.setViewport(vp);
        scrollPane.setViewportView(display);
        panel.add(scrollPane, BorderLayout.CENTER);

        // Create a speed slider for controling the speed of the animation.
        // Package the slider together with a label, and add it to the
        // overall display panel.
        JPanel sliderPanel = new JPanel();
        sliderPanel.setBorder(BorderFactory.createTitledBorder("Adjust Traversal Speed"));
        sliderPanel.add(new JLabel("Slow"));
        JSlider speedSlider = new JSlider(MIN_DELAY_MSECS, MAX_DELAY_MSECS,
                                          INITIAL_DELAY);
        speedSlider.setInverted(true);
        speedSlider.setPreferredSize(new Dimension(100, 
                                    speedSlider.getPreferredSize().height));
        speedSlider.setMaximumSize(speedSlider.getPreferredSize());
        speedSlider.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent evt) {
                timer.setDelay(((JSlider)evt.getSource()).getValue());
            }});
        sliderPanel.add(speedSlider);
        sliderPanel.add(new JLabel("Fast"));
        panel.add(sliderPanel, BorderLayout.SOUTH);

        // There are no actions defined for the graphics scrollpane, so
        // it is always "disabled."  The speed slider is always enabled.
        scrollPane.setEnabled(false);
        sliderPanel.setEnabled(true);

        return panel;
    }


    ////////////////////////
    // Action Definitions //
    ////////////////////////

    /** Animates a traversal algorithm.  It should be easy to visually
     *  verify a correct traversal algorithm.  This method will throw
     *  an exception if an incorrect traversal algorithm visits more
     *  locations than exist in the environment.
     *      @param  buttonLabel   name (label) of traversal algorithm
     *      @throws RuntimeException  if the algorithm visits more locations
     *                                than there are in the environment
     *                                (revisiting some)
     **/
    public void runTraversal(String buttonLabel)
    {
        // Find the appropriate iterator for the algorithm associated with
        // this button, which will be used to traverse through the
        // environment.
        enableControls(false);
        numLocationsVisited = 0;
        resetEnv();
        it = getIterator(buttonLabel, theEnv);
        timer.start();
    }
    
    /** Executes one step in a traversal. 
     **/
    public void step()
    {
        // Use the appropriate iterator to traverse through the environment,
        // placing color blocks as we go to animate the traversal algorithm.
        if ( it.hasNext() )
        {
            Location nextLoc = (Location)it.next();
            if ( theEnv.isEmpty(nextLoc) )
                new ColorBlock(highlightColor, theEnv, nextLoc);
            else
            {   // Be informative for pedagogical reasons.
                System.out.println("WARNING - TRAVERSAL BUG: " +
                    "revisiting location " + nextLoc);
                if ( numLocationsVisited >= numLocationsInEnv )
                    throw new RuntimeException("Too many iterations");
            }
        }
        else
            stopAlg();

        display.showEnv();
        numLocationsVisited++;
    }

    /** Interrupts a running traversal algorithm.
     **/
    public void stopAlg()
    {
        timer.stop();
        it = null;
        enableControls(true);
    }

    /** Resets the environment to be empty.
     **/
    public void resetEnv()
    {
        Iterator it = new RowMajorEnvIterator(theEnv);
        while ( it.hasNext() )
        {
            Locatable obj = theEnv.objectAt((Location)it.next());
            if ( obj != null )
                theEnv.remove(obj);
        }
        display.showEnv();
    }

    /** Constructs a new window in which to run another instance of
     *  the traversal algorithm animation program.
     **/
    public void newWindow()
    {
        BoundedEnv newEnv = 
                    new BoundedEnv(theEnv.numRows(), theEnv.numCols());
        TraversalGUI newFrame = 
                           new TraversalGUI(newEnv, bgColor, highlightColor,
                                            displayWidth, displayHeight);
        newFrame.setVisible(true);
    }

    /** Cleans up when a window is closing.
     */
    private void closingWindow()
    {
        openWindowCount--;
        if ( openWindowCount == 0 )
            System.exit(0);
    }

    /** Enables (or disables) the control panel buttons based on
     *  the program's state -- ready state or algorithm-running state.
     *     @param readyState    indicates whether the program is in
     *                          the ready state (true) or whether an
     *                          algorithm is currently running (readyState
     *                          is false)
     */
    private void enableControls(boolean readyState)
    {
        // Set the algorithm buttons, the reset button, and the
        // new window button to the readyState.
        Component[] internalComponents = algButtonList.getComponents();
        for ( int i = 0; i < internalComponents.length; i++ )
            internalComponents[i].setEnabled(readyState);
        resetButton.setEnabled(readyState);
        newWindowButton.setEnabled(readyState);

        // Stop button is on when everything else is off, and
        // off when everything else is on.
        stopButton.setEnabled(!readyState);
    }
    
    /** Nested class that is registered as the handler for exceptions
     *  on the Swing event thread. The handler will put up an alert panel,
     *  dump the stack trace to the console, and then exit entire program 
     *  rather than persist in an inconsistent state, which would be
     *  the default behavior.
     */
    public static class GUIExceptionHandler
    {
        public void handle(Throwable e)
        {
            e.printStackTrace();
            JOptionPane.showMessageDialog(null, 
                "Traversal program must exit due to an error:\n" + e +
                "\nSee System.out for more information.", 
                "Error", JOptionPane.ERROR_MESSAGE);
            System.exit(0);
        }
    }

}
